@@ -47,6 +47,8 @@ from b3j0f.utils.property import (
     put_properties, del_properties, get_local_property, get_property
 )
 
+from six import get_method_function
+
 from time import time
 
 try:
@@ -54,7 +56,7 @@ try:
 except ImportError:
     from dummy_threading import Timer
 
-from inspect import ismethod, getmembers
+from inspect import ismethod, getmembers, isfunction
 
 
 class Annotation(object):
@@ -74,9 +76,9 @@ class Annotation(object):
 
     And properties are:
     - propagate: (default True) determines if an annotation is propagated to
-        all sub target elements.
+    all sub target elements.
     - override: (default False) exclude previous annotation of the same type as
-        self class.
+    self class.
     - _ttl: (default None) self time to leave.
     - _in_memory: (default False) save instance in a global dictionary.
 
@@ -120,9 +122,9 @@ class Annotation(object):
     __ANNOTATIONS_IN_MEMORY__ = {}
 
     def __init__(
-        self,
-        on_bind_target=None, propagate=True, override=False, ttl=None,
-        in_memory=False
+            self,
+            on_bind_target=None, propagate=True, override=False, ttl=None,
+            in_memory=False
     ):
         """Default constructor with an 'on_bind_target' handler and propagate
         scope property.
@@ -147,7 +149,7 @@ class Annotation(object):
         self.ttl = ttl
         self.in_memory = in_memory
 
-        self.targets = set()
+        self.targets = list()
 
     def __call__(self, target, ctx=None):
         """Shouldn't be overriden by sub classes.
@@ -163,8 +165,7 @@ class Annotation(object):
         return result
 
     def __del__(self):
-        """Remove self to self.target annotations.
-        """
+        """Remove self to self.target annotations."""
 
         try:
             # nonify self ttl
@@ -175,6 +176,7 @@ class Annotation(object):
                 # remove self from target
                 try:
                     self.remove_from(target)
+
                 except TypeError:
                     # raised if target is not hashable
                     pass
@@ -189,7 +191,7 @@ class Annotation(object):
     def ttl(self):
         """Get actual ttl in seconds.
 
-        :return: actual ttl
+        :return: actual ttl.
         :rtype: float
         """
 
@@ -207,16 +209,18 @@ class Annotation(object):
     def ttl(self, value):
         """Change self ttl with input value.
 
-        :param float value: new ttl in seconds
+        :param float value: new ttl in seconds.
         """
 
         # get timer
         timer = getattr(self, Annotation.__TIMER, None)
+
         # if timer is running, stop the timer
         if timer is not None:
             timer.cancel()
+
         # initialize timestamp
-        ts = None
+        timestamp = None
 
         # if value is None
         if value is None:
@@ -224,15 +228,15 @@ class Annotation(object):
             timer = None
 
         else:  # else, renew a timer
-            # get ts
-            ts = time() + value
+            # get timestamp
+            timestamp = time() + value
             # start a new timer
             timer = Timer(value, self.__del__)
             timer.start()
             # set/update attributes
 
         setattr(self, Annotation.__TIMER, timer)
-        setattr(self, Annotation.__TS, ts)
+        setattr(self, Annotation.__TS, timestamp)
 
     @property
     def in_memory(self):
@@ -252,14 +256,16 @@ class Annotation(object):
     def in_memory(self, value):
         """Add or remove self from global memory.
 
-        :param bool value: if True(False) ensure self is(is not) in memory
+        :param bool value: if True(False) ensure self is(is not) in memory.
         """
 
         self_class = self.__class__
         memory = Annotation.__ANNOTATIONS_IN_MEMORY__
+
         if value:
             annotations_memory = memory.setdefault(self_class, set())
             annotations_memory.add(self)
+
         else:
             if self_class in memory:
                 annotations_memory = memory[self_class]
@@ -287,7 +293,7 @@ class Annotation(object):
     def _bind_target(self, target, ctx=None):
         """Method to override in order to specialize binding of target.
 
-        :param target: target to bind
+        :param target: target to bind.
         :param ctx: target ctx.
         :return: bound target.
         """
@@ -314,7 +320,8 @@ class Annotation(object):
         local_annotations.insert(0, self)
 
         # add target to self targets
-        self.targets.add(target)
+        if target not in self.targets:
+            self.targets.append(target)
 
         return result
 
@@ -341,7 +348,9 @@ class Annotation(object):
 
         try:
             # get local annotations
-            local_annotations = get_local_property(target, annotations_key)
+            local_annotations = get_local_property(
+                target, annotations_key, ctx=ctx
+            )
         except TypeError:
             raise TypeError('target {0} must be hashable'.format(target))
 
@@ -359,9 +368,8 @@ class Annotation(object):
                     del_properties(target, annotations_key)
 
     @classmethod
-    def free_memory(annotation_type, exclude=None):
-        """Free global annotation memory.
-        """
+    def free_memory(cls, exclude=None):
+        """Free global annotation memory."""
 
         annotations_in_memory = Annotation.__ANNOTATIONS_IN_MEMORY__
 
@@ -372,15 +380,15 @@ class Annotation(object):
             if issubclass(annotation_cls, exclude):
                 continue
 
-            if issubclass(annotation_cls, annotation_type):
+            if issubclass(annotation_cls, cls):
                 del annotations_in_memory[annotation_cls]
 
     @classmethod
-    def get_memory_annotations(annotation_type, exclude=None):
-        """Get annotations in memory which inherits from annotation_type.
+    def get_memory_annotations(cls, exclude=None):
+        """Get annotations in memory which inherits from cls.
 
-        :param tuple/type exclude: annotation type(s) to exclude from search
-        :return: found annotations which inherits from annotation_type.
+        :param tuple/type exclude: annotation type(s) to exclude from search.
+        :return: found annotations which inherits from cls.
         :rtype: set
         """
 
@@ -399,27 +407,27 @@ class Annotation(object):
                 continue
 
             # if annotation class inherits from self, add it in the result
-            if issubclass(annotation_cls, annotation_type):
+            if issubclass(annotation_cls, cls):
                 result |= annotations_in_memory[annotation_cls]
 
         return result
 
     @classmethod
     def get_local_annotations(
-        annotation_type, target, exclude=None, ctx=None, select=lambda *p: True
+            cls, target, exclude=None, ctx=None, select=lambda *p: True
     ):
         """Get a list of local target annotations in the order of their
             definition.
 
-        :param type annotation_type: type of annotation to get from target.
+        :param type cls: type of annotation to get from target.
         :param target: target from where get annotations.
         :param tuple/type exclude: annotation types to exclude from selection.
         :param ctx: target ctx.
         :param select: selection function which takes in parameters a target,
-        a ctx and an annotation and returns True if the annotation has to be
-        selected. True by default.
+            a ctx and an annotation and returns True if the annotation has to
+            be selected. True by default.
 
-        :return: target local annotations
+        :return: target local annotations.
         :rtype: list
         """
 
@@ -435,21 +443,28 @@ class Annotation(object):
             )
             if not local_annotations:
                 if ismethod(target):
+                    func = get_method_function(target)
                     local_annotations = get_local_property(
-                        target.__func__, Annotation.__ANNOTATIONS_KEY__,
+                        func, Annotation.__ANNOTATIONS_KEY__,
                         result, ctx=ctx
                     )
                     if not local_annotations:
                         local_annotations = get_local_property(
-                            target.__func__, Annotation.__ANNOTATIONS_KEY__,
+                            func, Annotation.__ANNOTATIONS_KEY__,
                             result
                         )
+                elif isfunction(target):
+                    local_annotations = get_local_property(
+                        target, Annotation.__ANNOTATIONS_KEY__,
+                        result
+                    )
+
         except TypeError:
             raise TypeError('target {0} must be hashable'.format(target))
 
         for local_annotation in local_annotations:
-            # check if local annotation inherits from annotation_type
-            inherited = isinstance(local_annotation, annotation_type)
+            # check if local annotation inherits from cls
+            inherited = isinstance(local_annotation, cls)
             # and if not excluded
             not_excluded = not isinstance(local_annotation, exclude)
             # and if selected
@@ -461,18 +476,16 @@ class Annotation(object):
         return result
 
     @classmethod
-    def remove(
-        annotation_type, target, exclude=None, ctx=None, select=lambda *p: True
-    ):
-        """Remove from target annotations which inherit from annotation_type
+    def remove(cls, target, exclude=None, ctx=None, select=lambda *p: True):
+        """Remove from target annotations which inherit from cls.
 
         :param target: target from where remove annotations which inherits from
-            annotation_type.
+            cls.
         :param tuple/type exclude: annotation types to exclude from selection.
         :param ctx: target ctx.
         :param select: annotation selection function which takes in parameters
-        a target, a ctx and an annotation and return True if the annotation has
-        to be removed.
+            a target, a ctx and an annotation and return True if the annotation
+            has to be removed.
         """
 
         # initialize exclude
@@ -483,18 +496,21 @@ class Annotation(object):
             local_annotations = get_local_property(
                 target, Annotation.__ANNOTATIONS_KEY__
             )
+
         except TypeError:
             raise TypeError('target {0} must be hashable'.format(target))
 
         # if there are local annotations
         if local_annotations is not None:
 
-            # get annotations to remove which inherits from annotation_type
+            # get annotations to remove which inherits from cls
             annotations_to_remove = [
                 annotation for annotation in local_annotations
-                if isinstance(annotation, annotation_type)
+                if (
+                    isinstance(annotation, cls)
                     and not isinstance(annotation, exclude)
                     and select(target, ctx, annotation)
+                )
             ]
 
             # and remove annotations from target
@@ -503,37 +519,55 @@ class Annotation(object):
 
     @classmethod
     def get_annotations(
-        annotation_type, target, exclude=None, ctx=None, select=lambda *p: True
+            cls, target,
+            exclude=None, ctx=None, select=lambda *p: True,
+            depth=1, followannotated=True, public=True, _history=None
     ):
-        """Returns all input target annotations of annotation_type type sorted
+        """Returns all input target annotations of cls type sorted
         by definition order.
 
-        :param type annotation_type: type of annotation to get from target.
+        :param type cls: type of annotation to get from target.
         :param target: target from where get annotations.
         :param tuple/type exclude: annotation types to remove from selection.
         :param ctx: target ctx.
         :param select: bool function which select annotations after applying
-        previous type filters. Takes a target, a ctx and an annotation in
-        parameters. True by default.
+            previous type filters. Takes a target, a ctx and an annotation in
+            parameters. True by default.
+        :param int depth: depth for searching annotations.
+        :param bool followannotated: if True (default) follow deeply only
+            annotated members.
+        :param bool public: if True (default) follow public members.
+        :param list _history: private parameter which save parsed elements.
+        :rtype: Annotation
         """
 
         result = []
 
-        annotations_by_ctx = get_property(
-            elt=target, key=Annotation.__ANNOTATIONS_KEY__, ctx=ctx
-        )
+        try:
+            annotations_by_ctx = get_property(
+                elt=target, key=Annotation.__ANNOTATIONS_KEY__, ctx=ctx
+            )
+
+        except TypeError:
+            annotations_by_ctx = {}
 
         if not annotations_by_ctx:
             if ismethod(target):
+                func = get_method_function(target)
                 annotations_by_ctx = get_property(
-                    elt=target.__func__,
+                    elt=func,
                     key=Annotation.__ANNOTATIONS_KEY__,
                     ctx=ctx
                 )
                 if not annotations_by_ctx:
                     annotations_by_ctx = get_property(
-                        elt=target.__func__, key=Annotation.__ANNOTATIONS_KEY__
+                        elt=func, key=Annotation.__ANNOTATIONS_KEY__
                     )
+            elif isfunction(target):
+                annotations_by_ctx = get_property(
+                    elt=target,
+                    key=Annotation.__ANNOTATIONS_KEY__
+                )
 
         exclude = () if exclude is None else exclude
 
@@ -554,38 +588,59 @@ class Annotation(object):
                     exclude += (annotation.__class__, )
 
                 # check for annotation
-                if (isinstance(annotation, annotation_type)
+                if (isinstance(annotation, cls)
                         and not isinstance(annotation, exclude)
                         and select(target, ctx, annotation)):
 
                     result.append(annotation)
 
+        if depth > 1 and (result or not followannotated):
+
+            if _history is None:
+                _history = [target]
+
+            else:
+                _history.append(target)
+
+            for name, member in getmembers(target):
+                if (name[0] != '_' or not public) and member not in _history:
+
+                    if ismethod(target) and name.startswith('im_'):
+                        continue
+
+                    result += cls.get_annotations(
+                        target=member, exclude=exclude, ctx=ctx, select=select,
+                        depth=depth-1, followannotated=followannotated,
+                        _history=_history
+                    )
+
         return result
 
     @classmethod
-    def get_annotated_fields(
-        annotation_type, instance, select=lambda *p: True
-    ):
-        """Get dict of {annotated fields: annotations} by annotation_type of
+    def get_annotated_fields(cls, instance, select=lambda *p: True):
+        """Get dict of {annotated fields: annotations} by cls of
         input instance.
 
-        :return: a set of (annotated fields, annotations)
+        :return: a set of (annotated fields, annotations).
         :rtype: dict
         """
 
         result = {}
 
-        for name, member in getmembers(instance):
+        for _, member in getmembers(instance):
             try:
-                annotations = annotation_type.get_annotations(
+                annotations = cls.get_annotations(
                     target=member, ctx=instance, select=select
                 )
+
             except TypeError:
                 pass
+
             else:
                 try:
                     if annotations:
                         result[member] = annotations
+
                 except TypeError:  # if field is an object proxy
                     pass
 
@@ -593,8 +648,7 @@ class Annotation(object):
 
 
 class StopPropagation(Annotation):
-    """Stop propagation for annotation types.
-    """
+    """Stop propagation for annotation types."""
 
     ANNOTATION_TYPES = 'annotation_types'
 
@@ -611,8 +665,7 @@ class StopPropagation(Annotation):
 
 class RoutineAnnotation(Annotation):
     """Dedicated to add information on any routine, routine parameters or
-    routine result.
-    """
+    routine result."""
 
     #: routine attribute name
     ROUTINE = 'routine'
@@ -626,12 +679,12 @@ class RoutineAnnotation(Annotation):
     __slots__ = (ROUTINE, RESULT, PARAMS) + Annotation.__slots__
 
     def __init__(
-        self, routine=None, params=None, result=None, *args, **kwargs
+            self, routine=None, params=None, result=None, *args, **kwargs
     ):
         """
-        :param routine: routine information
-        :param params: params information
-        :param result: routine result information
+        :param routine: routine information.
+        :param params: params information.
+        :param result: routine result information.
         """
 
         super(RoutineAnnotation, self).__init__(*args, **kwargs)
